@@ -3,7 +3,7 @@
  *
  * Copyright (C) 2010-2012  Tetsuo Handa <penguin-kernel@I-love.SAKURA.ne.jp>
  *
- * Version: 1.0.26   2012/04/01
+ * Version: 1.0.27   2012/05/05
  */
 
 #include "internal.h"
@@ -531,7 +531,26 @@ static int ccs_open(struct file *f)
 
 #endif
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
+
+/**
+ * ccs_file_open - Check permission for open().
+ *
+ * @f:    Pointer to "struct file".
+ * @cred: Pointer to "struct cred".
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_file_open(struct file *f, const struct cred *cred)
+{
+	int rc = ccs_open(f);
+	if (rc)
+		return rc;
+	while (!original_security_ops.file_open);
+	return original_security_ops.file_open(f, cred);
+}
+
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
 
 /**
  * ccs_dentry_open - Check permission for open().
@@ -634,7 +653,60 @@ static int ccs_inode_permission(struct inode *inode, int mask,
 
 #if defined(CONFIG_SECURITY_PATH)
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 33)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
+
+/**
+ * ccs_path_chown - Check permission for chown()/chgrp().
+ *
+ * @path:  Pointer to "struct path".
+ * @user:  User ID.
+ * @group: Group ID.
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_path_chown(struct path *path, kuid_t user, kgid_t group)
+{
+	int rc = ccs_chown_permission(path->dentry, path->mnt, user, group);
+	if (rc)
+		return rc;
+	while (!original_security_ops.path_chown);
+	return original_security_ops.path_chown(path, user, group);
+}
+
+/**
+ * ccs_path_chmod - Check permission for chmod().
+ *
+ * @path: Pointer to "struct path".
+ * @mode: Mode.
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_path_chmod(struct path *path, umode_t mode)
+{
+	int rc = ccs_chmod_permission(path->dentry, path->mnt, mode);
+	if (rc)
+		return rc;
+	while (!original_security_ops.path_chmod);
+	return original_security_ops.path_chmod(path, mode);
+}
+
+/**
+ * ccs_path_chroot - Check permission for chroot().
+ *
+ * @path: Pointer to "struct path".
+ *
+ * Returns 0 on success, negative value otherwise.
+ */
+static int ccs_path_chroot(struct path *path)
+{
+	int rc = ccs_chroot_permission(path);
+	if (rc)
+		return rc;
+	while (!original_security_ops.path_chroot);
+	return original_security_ops.path_chroot(path);
+}
+
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 33)
 
 /**
  * ccs_path_chown - Check permission for chown()/chgrp().
@@ -804,10 +876,17 @@ static int ccs_inode_setattr(struct dentry *dentry, struct iattr *attr)
 {
 	int rc = 0;
 #if !defined(CONFIG_SECURITY_PATH) || LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
+	if (attr->ia_valid & ATTR_UID)
+		rc = ccs_chown_permission(dentry, NULL, attr->ia_uid, INVALID_GID);
+	if (!rc && (attr->ia_valid & ATTR_GID))
+		rc = ccs_chown_permission(dentry, NULL, INVALID_UID, attr->ia_gid);
+#else
 	if (attr->ia_valid & ATTR_UID)
 		rc = ccs_chown_permission(dentry, NULL, attr->ia_uid, -1);
 	if (!rc && (attr->ia_valid & ATTR_GID))
 		rc = ccs_chown_permission(dentry, NULL, -1, attr->ia_gid);
+#endif
 	if (!rc && (attr->ia_valid & ATTR_MODE))
 		rc = ccs_chmod_permission(dentry, NULL, attr->ia_mode);
 #endif
@@ -2729,7 +2808,9 @@ static void __init ccs_update_security_ops(struct security_operations *ops)
 	swap_security_ops(bprm_committing_creds);
 #endif
 	/* Various permission checker. */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
+	swap_security_ops(file_open);
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
 	swap_security_ops(dentry_open);
 #else
 	swap_security_ops(inode_permission);
@@ -2808,7 +2889,7 @@ static int __init ccs_init(void)
 #endif
 	ccs_main_init();
 	ccs_update_security_ops(ops);
-	printk(KERN_INFO "AKARI: 1.0.26   2012/04/01\n");
+	printk(KERN_INFO "AKARI: 1.0.27   2012/05/05\n");
 	printk(KERN_INFO
 	       "Access Keeping And Regulating Instrument registered.\n");
 	return 0;
